/* 
        pknplay.cc

        MASL/M-MASL Copyright (c) KUNI/TDU$B1~>p8&(B
        Windows version Copyright (c) 1995, 1996 Tomoatsu Shimada  (shimarin)
        X Window System version Copyright (c)1997 Yamahata Isaku (ISA-9)
*/
#include <string.h>

#include "MaslInputStream.h"
#include "MmsConst.h"
#include "Pcm.h"
#include "printError.h"
#include "Scenario.h"

static void play (MaslInputStream& maslIs,const long endptr,
		  const char* const fileName,const Bool isOutputSjis);
static void divide (MaslInputStream& maslIs,const long endptr,
		    const Bool isOutputSjis);

main (int argc,char** argv)
{
  if (argc<2)
    {
      cerr << "Usage:" << argv[0] << " .pkn" << endl;
      exit (-1);
    }

  Bool isDivide=False;
  const char* fileName=argv[1];
  if (strcmp (argv[1],"-d")==0)
    {
      isDivide=True;
      fileName=argv[2];
    }

  const Bool isOutputSjis=False;
  //$B$3$3$GA3$k$Y$-%*%W%7%g%s$J$i$P!"(BTrue$B$K$9$k!#(B
  
  MaslInputStream maslIs (fileName);
  if (!maslIs)
    {
      printErrorCanNotOpen (fileName);
      maslIs.close ();
      exit (-1);
    }
  
  ///////////////////////////data size$B<hF@(B
  const unsigned int size=maslIs.getFileSize ();
  const long endptr=maslIs.tellg ()+size;//pkn data$B$N=*C<(B

  ///////////////PKN* $B$NFI$_9~$_(B
  char PKN[8];
  maslIs.read (PKN,8);
  PKN[3]='\0';
  if (strcmp (PKN,"PKN"))
    {
      cerr << "Incorrect data format detected" << endl;
      maslIs.close ();
      exit (-1);
    }
  
  ///////////////////////////////////////////////////////data$BFI$_9~$_(B
  char pcmAuther[33];
  pcmAuther[32]='\0';
  maslIs.read (pcmAuther,32);//$B$3$l$O2?$b$7$J$$(B
  cout << "the auther of this .pkn is ";
  Scenario::writeSjis (cout,pcmAuther,strlen (pcmAuther),isOutputSjis);
  cout << endl;
  cout << "now loading ";
  Scenario::writeSjis (cout,fileName,strlen (fileName),isOutputSjis);
  cout << "." << flush;
  
  if (!isDivide)
    {
      play (maslIs,endptr,fileName,isOutputSjis);
    }
  else
    {
      cout << endl;
      divide (maslIs,endptr,isOutputSjis);
    }
}


static void play (MaslInputStream& maslIs,const long endptr,
		  const char* const fileName,const Bool isOutputSjis)
{
  int numberOfPcm=0;
  Pcm* pPcm[MAXPCM];
  for (int i=0;i<MAXPCM;i++)
    pPcm[i]=NULL;
  while (maslIs.tellg ()<endptr)
    {
      unsigned char type;
      type=maslIs.get ();
      pPcm[numberOfPcm++]=new Pcm (type,maslIs);//pcm data$BFI$_9~$_(B
      cout << "." << flush;
    }
  cout << "done." << endl;
  Scenario::writeSjis (cout,fileName,strlen (fileName),isOutputSjis);
  cout << " has " << numberOfPcm << " PCM" << endl;
  maslIs.close ();
  
  //////////////////////////////////////////////////$B>pJsI=<((B
  cout << "No. \t format \t size  \t rate \t original data \t length " << endl;
  cout << "--- \t--------\t-------\t------\t---------------\t--------" << endl;
  for (int i=0;i<numberOfPcm;i++)
    {
      cout << i << "\t"
	   << ((
		(pPcm[i]->getType ())==1||(pPcm[i]->getType ()==3))?
	       "adpcm   ":"8bit    ")
	   << "\t"
	   << pPcm[i]->getSize () << "\t"
	   << pPcm[i]->getPcmRate () << "\t";
      const char* const pcmFileName=pPcm[i]->getFileName ();
      Scenario::writeSjis (cout,pcmFileName,strlen (pcmFileName),isOutputSjis);
      cout << "   \t"
	   << (float) pPcm[i]->getSize ()/ (float) pPcm[i]->getPcmRate ()
	   << endl;
    }

  //////////////////////////////////////////////////////pcm$B:F@8(B
  if (!Audio::openAudioDevice ())
    {
      perror ("can not open Audio Device");
      exit (-1);
    }
  Audio::initializeAudioDevice (DEFAULT_SAMPLE_RATE);
  for (int i=0;i<numberOfPcm;i++)
    {
      getchar ();
      const char* const pcmFileName=pPcm[i]->getFileName ();
      Scenario::writeSjis (cout,pcmFileName,strlen (pcmFileName),isOutputSjis);
      cout << ":now playing " << i << "-th pcm..." << flush;
      Audio::setRate (pPcm[i]->getPcmRate ());
      pPcm[i]->play ();
      Audio::sync ();
      cout << "done." << endl;
    }

  Audio::closeAudioDevice ();
  for (int i=0;i<MAXPCM;i++)
    delete pPcm[i];
}


static void divide (MaslInputStream& maslIs,const long endptr,
		    const Bool isOutputSjis)
{
  ////////////////////////////////////////////////////$B9=@.MWAG$KJ,3d(B
  while (maslIs.tellg ()<endptr)
    {
      unsigned char type;
      type=maslIs.get ();
      char pcmFileName[16];
      memset (pcmFileName,0,17);
      maslIs.read (pcmFileName,16);
      const int pcmRate=maslIs.get16 ();
      const int dataSize=maslIs.get32 ();
      char* const pcmData=new char[dataSize];
      maslIs.read (pcmData,dataSize);
      cout << "now creating \"";
      Scenario::writeSjis (cout,pcmFileName,strlen (pcmFileName),isOutputSjis);
      cout << "\"\tformat=" << ((type==1||type==3)?"adpcm":"8bit") << "\t"
	   << "size=" << dataSize << "byte\t"
	   << "rate=" << pcmRate << "hz\t" << endl;

      ofstream pcm(pcmFileName,ios::out|ios::noreplace);
      if (!pcm)
	{
	  printErrorCanNotOpen (pcmFileName);
	  pcm.close ();
	  delete[] pcmData;
	  continue;
	}

      pcm.write (pcmData,dataSize);
      if (pcm.fail ())
	{
	  printErrorCanNotWrite (pcmFileName);
	  pcm.close ();
	  delete[] pcmData;
	  continue;
      	}

      pcm.close ();
      delete[] pcmData;
    }

  maslIs.close ();//$B$J$<$+(Bsegmentation fault$B$9$k!#860x$OITL@!#(B
}
